from dune.perftool.ufl.modified_terminals import Restriction
from dune.perftool.generation import (iname,
                                      domain,
                                      get_global_context_value,
                                      temporary_variable,
                                      instruction)
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           name_localcenter,
                                           pymbolic_in_cell_coordinates)

from dune.perftool.pdelab.quadrature import quadrature_inames
from dune.perftool.generation.counter import get_counted_variable
from dune.perftool.options import get_form_option
import pymbolic.primitives as prim


# add inames over the micro elements in tensor representation,
# i.e. each element has (i_1,i_2,...,i_d) indices
@iname
def sub_element_inames():
    name = "subel"
    dim = local_dimension()
    inames = tuple()
    for i in range(dim):
        name_counted = get_counted_variable(name)
        inames = inames + (name_counted,)
        domain(name_counted, get_form_option("number_of_blocks"))
    return inames


# define inames for boundary integration
# In the case of boundary integrationsince we have only d-1 loops,
# but we need always d inames to compute the macro index.
# Therefore we must find out which iname is constant.
# Since loo.py cannot handle if-else blocks very well, this whole
# computation seems a bit cumbersome.
def sub_facet_inames():
    subelem_inames = sub_element_inames()

    center = pymbolic_in_cell_coordinates(prim.Variable(name_localcenter()), Restriction.NEGATIVE)

    # check if iname[index] must be constant or not
    def predicate(index):
        return prim.Comparison(prim.Subscript(center, index), "==", 0.5)

    def conditional_instruction(index):
        # instruction for not constant iname
        # special case for the third iname
        instruction(assignee=prim.Variable(inames[index]),
                    expression=prim.Variable(subelem_inames[1 if index == 2 else 0]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([predicate(index)])
                    )
        # instruction for constant iname
        instruction(assignee=prim.Variable(inames[index]),
                    expression=prim.Product(((k - 1), prim.Subscript(center, (index,)))),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalNot(predicate(index))])
                    )

    k = get_form_option("number_of_blocks")

    inames = ("x",)
    temporary_variable(inames[0])
    conditional_instruction(0)
    if world_dimension() < 3:
        inames = inames + ("y",)
        temporary_variable(inames[1])
        conditional_instruction(1)
    else:
        inames = inames + ("y",)
        temporary_variable(inames[1])
        # one additional condition is needed in 3d for the second iname
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Variable(subelem_inames[0]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalAnd((predicate(1), predicate(2)))])
                    )
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Variable(subelem_inames[1]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalAnd((predicate(1), predicate(0)))])
                    )
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Product(((k - 1), prim.Subscript(center, (1,)))),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalNot(predicate(1))])
                    )

        inames = inames + ("z",)
        temporary_variable(inames[2])
        conditional_instruction(2)

    return inames


# compute sequential index for given tensor index, the same as index in base-k to base-10
def tensor_index_to_sequential_index(indices, k):
    return prim.Sum(tuple(index * k ** i for i, index in enumerate(indices)))


# compute tensor index for given sequential index, the same as index in base-10 to base-k
def to_tensor_index(iname, k):
    return tuple(prim.Remainder(prim.Variable(iname) / (k**i), k) for i in range(world_dimension()))


# compute index for higher order FEM for a given Q1 index
def micro_index_to_macro_index(element, iname):
    it = get_global_context_value("integral_type")
    if it == "cell":
        subelem_inames = sub_element_inames()
    elif it == "exterior_facet" or it == "interior_facet":
        subelem_inames = sub_facet_inames()

    k = get_form_option("number_of_blocks")
    p = element.degree()
    modified_index = prim.Sum((tensor_index_to_sequential_index(to_tensor_index(iname, p + 1), p * k + 1),
                               tensor_index_to_sequential_index(tuple(prim.Variable(iname) * p for iname in subelem_inames), p * k + 1)))

    return modified_index
