"""
Implement tools around the idea of having two 'views' on memory:
One being an ordinary array with proper shape and so on, and one
being a an array of SIMD vectors
"""

from dune.perftool.loopy.target import dtype_floatingpoint
from dune.perftool.loopy.vcl import get_vcl_type_size

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
import pytools as pt


def get_vector_view_name(tmpname):
    return tmpname + "_vec"


def add_vector_view(knl, tmpname, pad_to=None, flatview=False):
    """
    Kernel transformation to add a vector view temporary
    that interprets the same memory as another temporary
    """
    temporaries = knl.temporary_variables
    assert tmpname in temporaries
    temp = temporaries[tmpname]
    vecname = get_vector_view_name(tmpname)
    bsname = tmpname + "_base"

    if vecname in knl.temporary_variables:
        return knl

    # Add base storage to the original temporary!
    if not temp.base_storage:
        temp = temp.copy(base_storage=bsname,
                         _base_storage_access_may_be_aliasing=True,
                         )
        temporaries[tmpname] = temp
    else:
        bsname = temp.base_storage

    # Determine the shape by dividing total size by vector size
    # Also apply the padding we need for rotation
    # TODO: *Only* apply this padding if really needed (a bit hard to figure out)
    vecsize = get_vcl_type_size(temp.dtype)
    if all(isinstance(s, int) for s in temp.shape):
        size = pt.product(temp.shape) // vecsize
        if size % vecsize != 0:
            size = (size // vecsize + 1) * vecsize
    else:
        size = prim.FloorDiv(prim.Product(temp.shape), vecsize)
        size = (size // vecsize + 1) * vecsize

    # Maybe do some padding.
    if pad_to:
        size = (size // pad_to + 1) * pad_to

    # Some vectorview are intentionally flat! (e.g. the output buffers of
    # sum factorization kernels
    if flatview:
        shape = (size, vecsize)
        dim_tags = "c,vec"
    else:
        shape = temp.shape
        # This works around a loopy weirdness (which might as well be a bug)
        # TODO: investigate this!
        if len(shape) == 1:
            shape = (1, vecsize)
            dim_tags = "c,vec"
        else:
            dim_tags = temp.dim_tags[:-1] + ("vec",)

    # Now add a vector view temporary
    vecname = tmpname + "_vec"
    temporaries[vecname] = lp.TemporaryVariable(vecname,
                                                dim_tags=dim_tags,
                                                shape=shape,
                                                base_storage=bsname,
                                                dtype=dtype_floatingpoint(),
                                                scope=lp.temp_var_scope.PRIVATE,
                                                _base_storage_access_may_be_aliasing=True,
                                                )

    # Avoid that any of these temporaries are eliminated
    silenced = ['temp_to_write({})'.format(tmpname),
                'temp_to_write({})'.format(vecname),
                'read_no_write({})'.format(tmpname),
                'read_no_write({})'.format(vecname),
                ]

    return knl.copy(temporary_variables=temporaries,
                    silenced_warnings=knl.silenced_warnings + silenced)


def add_temporary_with_vector_view(knl, name, *args, **kwargs):
    temps = knl.temporary_variables
    assert name not in temps
    temps[name] = lp.TemporaryVariable(name, *args, **kwargs)
    knl = knl.copy(temporary_variables=temps)
    knl = add_vector_view(knl, name)
    return knl
