""" Manage the command line options to the form compiler executable """

from argparse import ArgumentParser
from os.path import abspath
from pytools import ImmutableRecord, memoize

from dune.testtools.parametertree.parser import parse_ini_file


class PerftoolOption(ImmutableRecord):
    """ Data structure representing a single formcompiler option """
    def __init__(self,
                 default=None,
                 helpstr="Undocumented feature!",
                 process=lambda x: x,
                 _type=type(None),
                 ):
        _type = type(default)
        if issubclass(_type, type(None)):
            _type = str
        ImmutableRecord.__init__(self,
                                 helpstr=helpstr,
                                 default=default,
                                 type=_type,
                                 )


class PerftoolGlobalOptionsArray(ImmutableRecord):
    """ A collection of form compiler arguments """
    def __init__(self, **kwargs):
        opts = {k: v.default for k, v in PerftoolGlobalOptionsArray.__dict__.items() if isinstance(v, PerftoolOption)}
        opts.update(**kwargs)
        ImmutableRecord.__init__(self, **opts)

    # Arguments that are to be set from the outside
    uflfile = PerftoolOption(helpstr="the UFL file to compile")
    debug_cache_with_stack = PerftoolOption(default=False, helpstr="Store stack along with cache objects. Makes debugging caching issues easier.")
    driver_file = PerftoolOption(helpstr="The filename for the generated driver header")
    explicit_time_stepping = PerftoolOption(default=False, helpstr="use explicit time stepping")
    exact_solution_expression = PerftoolOption(helpstr="name of the exact solution expression in the ufl file")
    compare_l2errorsquared = PerftoolOption(helpstr="maximal allowed l2 error squared of difference between numerical solution and interpolation of exact solution (NOTE: requires --exact-solution-expression)")
    l2error_tree_path = PerftoolOption(default=None, helpstr="Tree pathes that should be considered for l2 error calculation. Default None means we take all of them into account.")
    ini_file = PerftoolOption(helpstr="An inifile to use. A generated driver will be hard-coded to it, a [formcompiler] section will be used as default values to form compiler arguments (use snake case)")
    opcounter = PerftoolOption(default=False, helpstr="Count operations. Note: In this case only operator applications are generated since solving and operator counting does not work. You probably want to set instrumentation level>0.")
    performance_measuring = PerftoolOption(default=False, helpstr="Generate opcounter codepath, but only measure times!")
    instrumentation_level = PerftoolOption(default=0, helpstr="Control time/opcounter measurements. 0-do nothing, 1-measure program as a whole, 2-operator applications, 3-measure kernel (eg. alpha-volume, ...), 4-parts of kernel (eg. stage 1-3 of SF)")
    project_basedir = PerftoolOption(helpstr="The base (build) directory of the dune-perftool project")
    architecture = PerftoolOption(default="haswell", helpstr="The architecture to optimize for. Possible values: haswell|knl|skylake")
    yaspgrid_offset = PerftoolOption(default=False, helpstr="Set to true if you want a yasp grid where the lower left corner is not in the origin.")
    precision_bits = PerftoolOption(default=64, helpstr="The number of bits for the floating point type")
    overlapping = PerftoolOption(default=False, helpstr="Use an overlapping solver and constraints. You still need to make sure to construct a grid with overlap! The parallel option will be set automatically.")
    operators = PerftoolOption(default="r", helpstr="A comma separated list of operators, each name will be interpreted as a subsection name within the formcompiler section")
    target_name = PerftoolOption(default=None, helpstr="The target name from CMake")

    # Arguments that are mainly to be set by logic depending on other options
    max_vector_width = PerftoolOption(default=256, helpstr=None)
    parallel = PerftoolOption(default=False, helpstr="Mark that this program should be run in parallel. If set to true the c++ code will check that there are more than 1 MPI-ranks involved and the error computation will use communication.")


class PerftoolFormOptionsArray(ImmutableRecord):
    """ A collection of form-specific form compiler arguments """
    def __init__(self, **kwargs):
        opts = {k: v.default for k, v in PerftoolFormOptionsArray.__dict__.items() if isinstance(v, PerftoolOption)}
        opts.update(**kwargs)
        ImmutableRecord.__init__(self, **opts)

    # Form specific options
    form = PerftoolOption(default=None, helpstr="The name of the UFL object representing the form in the UFL file")
    filename = PerftoolOption(default=None, helpstr="The filename to use for this LocalOperator")
    classname = PerftoolOption(default=None, helpstr="The name of the C++ class to generate")
    numerical_jacobian = PerftoolOption(default=False, helpstr="use numerical jacobians (only makes sense, if uflpdelab for some reason fails to generate analytic jacobians)")
    matrix_free = PerftoolOption(default=False, helpstr="Use iterative solver with matrix free jacobian application")
    print_transformations = PerftoolOption(default=False, helpstr="print out dot files after ufl tree transformations")
    print_transformations_dir = PerftoolOption(default=".", helpstr="place where to put dot files (can be omitted)")
    quadrature_order = PerftoolOption(_type=int, helpstr="Quadrature order used for all integrals.")
    diagonal_transformation_matrix = PerftoolOption(default=False, helpstr="set option if the jacobian of the transformation is diagonal (axiparallel grids)")
    constant_transformation_matrix = PerftoolOption(default=False, helpstr="set option if the jacobian of the transformation is constant on a cell")
    fastdg = PerftoolOption(default=False, helpstr="Use FastDGGridOperator from PDELab.")
    sumfact = PerftoolOption(default=False, helpstr="Use sumfactorization")
    vectorization_quadloop = PerftoolOption(default=False, helpstr="whether to generate code with explicit vectorization")
    vectorization_strategy = PerftoolOption(default="none", helpstr="The identifier of the vectorization cost model. Possible values: none|explicit|model|fromlist")
    vectorization_horizontal = PerftoolOption(default=None, helpstr="an explicit value for horizontal vectorization read by the 'explicit' strategy")
    vectorization_vertical = PerftoolOption(default=None, helpstr="an explicit value for vertical vectorization read by the 'explicit' strategy")
    vectorization_padding = PerftoolOption(default=None, helpstr="an explicit value for the allowed padding in vectorization")
    vectorization_allow_quadrature_changes = PerftoolOption(default=False, helpstr="whether the vectorization strategy is allowed to alter quadrature point numbers")
    vectorization_list_index = PerftoolOption(default=None, helpstr="Which vectorization to pick from a list (only valid with vectorization_strategy=fromlist).")
    simplify = PerftoolOption(default=False, helpstr="Whether to simplify expressions using sympy")
    assure_statement_ordering = PerftoolOption(default=False, helpstr="Whether special care should be taken for a good statement ordering in sumfact kernels, runs into a loopy scheduler performance bug, but is necessary for production.")
    generate_jacobians = PerftoolOption(default=True, helpstr="Whether jacobian_* methods should be generated. This is set to false automatically, when numerical_jacobian is set to true.")
    unroll_dimension_loops = PerftoolOption(default=False, helpstr="whether loops over the geometric dimension should be unrolled")
    precompute_quadrature_info = PerftoolOption(default=True, helpstr="compute quadrature points and weights in the constructor of the local operator")
    blockstructured = PerftoolOption(default=False, helpstr="Use block structure")
    number_of_blocks = PerftoolOption(default=1, helpstr="Number of sub blocks in one direction")


# Until more sophisticated logic is needed, we keep the actual option data in this module
_global_options = PerftoolGlobalOptionsArray()
_form_options = {}


def initialize_options():
    """ Initialize the options from the command line """
    global _global_options
    _global_options = update_options_from_commandline(_global_options)
    _global_options = update_options_from_inifile(_global_options)


def update_options_from_commandline(opt):
    """ Return an options array object with updated values from the commandline """
    assert isinstance(opt, PerftoolGlobalOptionsArray)
    parser = ArgumentParser(description="Compile UFL files to PDELab C++ code",
                            epilog="Please report bugs to dominic.kempf@iwr.uni-heidelberg.de",
                            )
    parser.add_argument('--version', action='version', version='%(prog)s 0.1')
    for k, v in type(opt).__dict__.items():
        if isinstance(v, PerftoolOption) and v.helpstr is not None:
            cmdopt = "--{}".format(k.replace('_', '-'))
            parser.add_argument(cmdopt, help=v.helpstr, type=v.type)
    parsedargs = {k: v for k, v in vars(parser.parse_args()).items() if v is not None}
    return opt.copy(**parsedargs)


def update_options_from_inifile(opt):
    """ Return an options array object with updated values from an inifile """
    if opt.ini_file:
        def parse_ini(section, opttype):
            def _fix_types(k, v):
                if hasattr(opttype, k) and getattr(opttype, k).type is bool:
                    return bool(eval(v))
                if hasattr(opttype, k):
                    return getattr(opttype, k).type(v)
                return v
            ini = parse_ini_file(opt.ini_file).get(section, {})
            return {k: _fix_types(k, v) for k, v in ini.items()}

        opt = opt.copy(**parse_ini("formcompiler", PerftoolGlobalOptionsArray))
        # Also parse form-specific options
        for form in [i.strip() for i in opt.operators.split(",")]:
            _form_options[form] = PerftoolFormOptionsArray(**parse_ini("formcompiler.{}".format(form), PerftoolFormOptionsArray))

    return opt


@memoize
def process_global_options(opt):
    """ Make sure that the options have been fully processed """
    opt = expand_architecture_options(opt)

    if opt.overlapping:
        opt = opt.copy(parallel=True)

    return opt


@memoize
def process_form_options(opt, form):
    if opt.sumfact:
        opt = opt.copy(unroll_dimension_loops=True)

    if opt.numerical_jacobian:
        opt = opt.copy(generate_jacobians=False)

    if opt.form is None:
        opt = opt.copy(form=form)

    if opt.classname is None:
        opt = opt.copy(classname="{}Operator".format(opt.form))

    if opt.filename is None:
        opt = opt.copy(filename="{}_{}_file.hh".format(get_option("target_name"), opt.classname))

    return opt


def expand_architecture_options(opt):
    if opt.architecture == "haswell":
        return opt.copy(max_vector_width=256)
    elif opt.architecture == "knl":
        return opt.copy(max_vector_width=512)
    elif opt.architecture == "skylake":
        return opt.copy(max_vector_width=512)
    else:
        raise NotImplementedError("Architecture {} not known!".format(opt.architecture))


def set_option(key, value):
    """Add the key value pair to the options.

    If the key is already in the options dictionary its value will be
    overwritten.  Form compiler arguments will always be set before
    any other options.
    """
    global _global_options
    _global_options = process_global_options(_global_options).copy(**{key: value})


def set_form_option(key, value, form=None):
    if form is None:
        from dune.perftool.generation import get_global_context_value
        form = get_global_context_value("form_identifier", 0)
    if isinstance(form, int):
        form = get_option("operators").split(",")[form].strip()
    _form_options[form] = _form_options[form].copy(**{key: value})


def get_option(key):
    processed_global_opts = process_global_options(_global_options)
    return getattr(processed_global_opts, key)


def get_form_option(key, form=None):
    if form is None:
        from dune.perftool.generation import get_global_context_value
        form = get_global_context_value("form_identifier", 0)
    if isinstance(form, int):
        form = get_option("operators").split(",")[form].strip()
    processed_form_opts = process_form_options(_form_options[form], form)
    return getattr(processed_form_opts, key)


def option_switch(opt):
    def _switch():
        try:
            if get_option(opt):
                return opt
            else:
                return "default"
        except AttributeError:
            if get_form_option(opt):
                return opt
            else:
                return "default"
    return _switch
