""" Timing related generator functions """

from dune.perftool.options import get_option
from dune.perftool.generation import (cached,
                                      include_file,
                                      pre_include,
                                      post_include,
                                      preamble,
                                      )
from dune.perftool.pdelab.driver import (get_form_ident,
                                         name_initree,
                                         name_mpihelper,
                                         )
from dune.perftool.pdelab.driver.gridfunctionspace import (name_leafview,
                                                           name_trial_gfs,
                                                           type_leafview,
                                                           )
from dune.perftool.pdelab.driver.gridoperator import (name_gridoperator,
                                                      name_localoperator,
                                                      type_gridoperator,
                                                      )
from dune.perftool.pdelab.driver.solve import (name_vector,
                                               type_vector,
                                               )


@preamble(section="timings")
def define_timing_identifier(name):
    ini = name_initree()
    return "auto {} = {}.get<std::string>(\"identifier\", std::string(argv[0]));".format(name, ini)


def name_timing_identifier():
    name = "ident"
    define_timing_identifier(name)
    return name


@preamble(section="timings")
def dump_dof_numbers(stream):
    ident = name_timing_identifier()
    level = get_option("instrumentation_level")
    include_file("dune/pdelab/common/partitionviewentityset.hh", filetag="driver")
    gvt = type_leafview()
    gv = name_leafview()
    return ["Dune::PDELab::NonOverlappingEntitySet<{}> es({});".format(gvt, gv),
            "{} << \"{} \" << {} << \" dofs dofs \" << {}.maxLocalSize() * es.size(0) << std::endl;".format(stream,
                                                                                                            level,
                                                                                                            ident,
                                                                                                            name_trial_gfs())
            ]


@preamble(section="timings")
def define_timing_stream(name):
    include_file('fstream', filetag='driver', system=True)
    include_file('sstream', filetag='driver', system=True)
    include_file('sys/types.h', filetag='driver', system=True)
    include_file('chrono', filetag='driver', system=True)

    return ["std::stringstream ss;",
            "ss << \"{}/timings-rank-\" << {}.rank() << \"-\" << std::chrono::high_resolution_clock::now().time_since_epoch().count() << \".csv\";".format(get_option('project_basedir'), name_mpihelper()),
            "std::ofstream {};".format(name),
            "{}.open(ss.str(), std::ios_base::app);".format(name),
            ]


def name_timing_stream():
    name = "timestream"
    define_timing_stream(name)
    dump_dof_numbers(name)
    return name


@cached
def setup_timer():
    # TODO check that we are using YASP?
    if get_option('opcounter'):
        pre_include("#define ENABLE_COUNTER", filetag="driver")
    pre_include("#define ENABLE_HP_TIMERS", filetag="driver")
    include_file("dune/perftool/common/timer.hh", filetag="driver")


@preamble(section="timings")
def evaluate_residual_timer():
    n_go = name_gridoperator(get_form_ident())
    v = name_vector(get_form_ident())
    t_v = type_vector(get_form_ident())
    setup_timer()

    if get_option('instrumentation_level') >= 2:
        # Write back times
        from dune.perftool.generation import post_include
        post_include("HP_DECLARE_TIMER(residual_evaluation);", filetag="driver")
        timestream = name_timing_stream()
        print_times = []

    lop_name = name_localoperator(get_form_ident())
    if get_option('instrumentation_level') >= 3:
        print_times.append("{}.dump_timers({}, {}, true);".format(lop_name, timestream, name_timing_identifier()))

    if get_option('instrumentation_level') >= 2:
        evaluation = ["HP_TIMER_START(residual_evaluation);",
                      "{}.residual({}, r);".format(n_go, v),
                      "HP_TIMER_STOP(residual_evaluation);",
                      "DUMP_TIMER({}, residual_evaluation, {}, true);".format(get_option("instrumentation_level"), timestream)]
        evaluation.extend(print_times)
    else:
        evaluation = ["{}.residual({}, r);".format(n_go, v)]

    evaluation = ["{} r({});".format(t_v, v), "r=0.0;"] + evaluation

    return evaluation


@preamble(section="timings")
def apply_jacobian_timer():
    n_go = name_gridoperator(get_form_ident())
    v = name_vector(get_form_ident())
    t_v = type_vector(get_form_ident())
    setup_timer()

    if get_option('instrumentation_level') >= 2:
        # Write back times
        from dune.perftool.generation import post_include
        post_include("HP_DECLARE_TIMER(apply_jacobian);", filetag="driver")
        timestream = name_timing_stream()
        print_times = []

    lop_name = name_localoperator(get_form_ident())
    if get_option('instrumentation_level') >= 3:
        print_times.append("{}.dump_timers({}, {}, true);".format(lop_name, timestream, name_timing_identifier()))

    if get_option('instrumentation_level') >= 2:
        evaluation = ["HP_TIMER_START(apply_jacobian);",
                      "{}.jacobian_apply({}, j);".format(n_go, v),
                      "HP_TIMER_STOP(apply_jacobian);",
                      "DUMP_TIMER({}, apply_jacobian, {}, true);".format(get_option("instrumentation_level"), timestream)]
        evaluation.extend(print_times)
    else:
        evaluation = ["{}.jacobian_apply({}, j);".format(n_go, v)]

    evaluation = ["{} j({});".format(t_v, v), "j=0.0;"] + evaluation

    return evaluation


@preamble(section="timings")
def assemble_matrix_timer():
    t_go = type_gridoperator(get_form_ident())
    n_go = name_gridoperator(get_form_ident())
    v = name_vector(get_form_ident())
    t_v = type_vector(get_form_ident())
    setup_timer()

    if get_option('instrumentation_level') >= 2:
        # Write back times
        from dune.perftool.generation import post_include
        post_include("HP_DECLARE_TIMER(matrix_assembly);", filetag="driver")
        timestream = name_timing_stream()
        print_times = []

    lop_name = name_localoperator(get_form_ident())
    if get_option('instrumentation_level') >= 3:
        print_times.append("{}.dump_timers({}, {}, true);".format(lop_name, timestream, name_timing_identifier()))

    if get_option('instrumentation_level') >= 2:
        assembly = ["HP_TIMER_START(matrix_assembly);",
                    "{}.jacobian({},m);".format(n_go, v),
                    "HP_TIMER_STOP(matrix_assembly);",
                    "DUMP_TIMER({}, matrix_assembly, {}, true);".format(get_option("instrumentation_level"), timestream)]
        assembly.extend(print_times)
    else:
        assembly = ["{}.jacobian({},m);".format(n_go, v)]

    assembly = ["using M = typename {}::Traits::Jacobian;".format(t_go),
                "M m({});".format(n_go)] + assembly

    return assembly
