""" Permute sum factorization kernels """

import itertools


def sumfact_permutation_heuristic(permutations, stage):
    """Heuristic to choose a permutation

    - Stage 1: Pick the permutation where in permutations[1:] most
      elements are ordered by size
    - Stage 3: Pick the permutation where in permutations[:-1] most
      elements are ordered by size
    """
    def cost(perm, stage):
        cost = 0
        for i in range(0, len(perm) - 2):
            if stage == 1:
                if perm[i + 1] > perm[i + 2]:
                    cost += 1
            if stage == 3:
                if perm[0] > perm[i + 1]:
                    cost += 1
        return cost

    perm = min(permutations, key=lambda i: cost(i, stage))
    return perm


def flop_cost(matrix_sequence):
    """Computational cost of sumfactorization with this matrix_sequence
    """
    cost = 0
    for l in range(len(matrix_sequence)):
        cost_m = 1
        cost_n = 1
        for i in range(l + 1):
            cost_m *= matrix_sequence[i].rows
        for i in range(l, len(matrix_sequence)):
            cost_n *= matrix_sequence[i].cols
        cost += cost_m * cost_n
    # The factor of 2 indicates FMA
    return 2 * cost


def sumfact_permutation_strategy(sf):
    """Choose permutation of the matrix sequence based on computational cost

    Note: If there are multiple permutations with the same cost a
    heuristic is used to pick one.
    """
    # Extract information from the SumfactKernel object
    matrix_sequence = sf.matrix_sequence
    stage = sf.stage

    # Combine permutation and matrix_sequence
    perm = [i for i, _ in enumerate(matrix_sequence)]
    perm_matrix_sequence = zip(perm, matrix_sequence)

    # Find cost for all possible permutations of the matrix_sequence
    perm_cost = []
    for permutation in itertools.permutations(perm_matrix_sequence):
        perm, series = zip(*permutation)
        cost = flop_cost(series)
        perm_cost.append((perm, cost))

    # Find minimal cost and all permutations with that cost
    _, costs = zip(*perm_cost)
    minimal_cost = min(costs)
    minimal_cost_permutations = [p[0] for p in perm_cost if p[1] == minimal_cost]

    # Use heuristic to pick one of the minimal cost permutations
    perm = sumfact_permutation_heuristic(minimal_cost_permutations, stage)
    return perm


def permute_forward(t, perm):
    tmp = []
    for pos in perm:
        tmp.append(t[pos])
    return tuple(tmp)


def permute_backward(t, perm):
    tmp = [None] * len(t)
    for i, pos in enumerate(perm):
        tmp[pos] = t[i]
    return tuple(tmp)
