from dune.perftool.generation import (backend,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      loopy_class_member,
                                      temporary_variable,
                                      )

from dune.perftool.sumfact.tabulation import (quadrature_points_per_direction,
                                              local_quadrature_points_per_direction,
                                              name_oned_quadrature_points,
                                              name_oned_quadrature_weights,
                                              )
from dune.perftool.pdelab.argument import name_accumulation_variable
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.options import get_form_option
from dune.perftool.sumfact.switch import get_facedir
from dune.perftool.loopy.target import dtype_floatingpoint

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def nest_quadrature_loops(kernel, inames):
    insns = []
    for insn in kernel.instructions:
        if "quad" in insn.tags:
            insns.append(insn.copy(within_inames=insn.within_inames.union(frozenset(inames)),
                                   tags=insn.tags - frozenset({"quad"})
                                   )
                         )
        else:
            insns.append(insn)

    return kernel.copy(instructions=insns)


class BaseWeight(FunctionIdentifier):
    def __init__(self, accumobj):
        self.accumobj = accumobj

    def __getinitargs__(self):
        return (self.accumobj,)

    @property
    def name(self):
        return '{}.weight'.format(self.accumobj)


@function_mangler
def base_weight_function_mangler(target, func, dtypes):
    if isinstance(func, BaseWeight):
        return CallMangleInfo(func.name, (NumpyType(dtype_floatingpoint()),), ())


def pymbolic_base_weight():
    """ This is the base weight that should be multiplied to the quadrature
    weight. With the fast DG assembler this will handle the weighting of the
    time discretization scheme.
    TODO: Introduce backend switch that uses above BaseWeight function
    """
    return 1.0


@backend(interface="quad_inames", name="sumfact")
@kernel_cached
def quadrature_inames(element):
    if element is None:
        names = tuple("quad_{}".format(d) for d in range(local_dimension()))
    else:
        from ufl import FiniteElement, TensorProductElement
        assert isinstance(element, (FiniteElement, TensorProductElement))
        from dune.perftool.pdelab.driver import FEM_name_mangling
        names = tuple("quad_{}_{}".format(FEM_name_mangling(element), d) for d in range(local_dimension()))

    local_qps_per_dir = local_quadrature_points_per_direction()
    domain(names, local_qps_per_dir)
    return names


@iname(kernel="operator")
def constructor_quad_iname(name, d, bound):
    name = "{}_localdim{}".format(name, d)
    domain(name, bound, kernel="operator")
    return name


def constructor_quadrature_inames(dim):
    local_qps_per_dir = local_quadrature_points_per_direction()
    local_qps_per_dir_str = '_'.join(map(str, local_qps_per_dir))
    name = "quadiname_dim{}_num{}".format(dim, local_qps_per_dir_str)
    return tuple(constructor_quad_iname(name, d, local_qps_per_dir[d]) for d in range(local_dimension()))


def define_recursive_quadrature_weight(visitor, name, direction):
    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    iname = quadrature_inames(element)[direction]
    temporary_variable(name, shape=(), shape_impl=())

    qps_per_dir = quadrature_points_per_direction()
    qp_bound = qps_per_dir[direction]
    instruction(expression=Product((recursive_quadrature_weight(direction + 1),
                                    Subscript(Variable(name_oned_quadrature_weights(qp_bound)),
                                              (Variable(iname),)
                                              ))
                                   ),
                assignee=Variable(name),
                forced_iname_deps=frozenset(quadrature_inames()[direction:]),
                forced_iname_deps_is_final=True,
                tags=frozenset({"quad"}),
                )


def recursive_quadrature_weight(visitor, direction=0):
    if direction == local_dimension():
        return pymbolic_base_weight()
    else:
        name = 'weight_{}'.format(direction)
        define_recursive_quadrature_weight(visitor, name, direction)
        return Variable(name)


def quadrature_weight(visitor):
    # Return non-precomputed version
    if not get_form_option("precompute_quadrature_info"):
        return recursive_quadrature_weight(visitor)

    # Quadrature points per (local) direction
    dim = local_dimension()
    local_qps_per_dir = local_quadrature_points_per_direction()
    local_qps_per_dir_str = '_'.join(map(str, local_qps_per_dir))
    name = "quad_weights_dim{}_num{}".format(dim, local_qps_per_dir_str)

    # Add a class member
    loopy_class_member(name,
                       shape=local_qps_per_dir,
                       classtag="operator",
                       dim_tags=",".join(["f"] * dim),
                       managed=True,
                       potentially_vectorized=True,
                       )

    # Precompute it in the constructor
    inames = constructor_quadrature_inames(dim)
    assignee = prim.Subscript(prim.Variable(name), tuple(prim.Variable(i) for i in inames))
    expression = prim.Product(tuple(Subscript(Variable(name_oned_quadrature_weights(local_qps_per_dir[index])),
                                              (prim.Variable(iname),)) for index, iname in enumerate(inames)))
    instruction(assignee=assignee,
                expression=expression,
                within_inames=frozenset(inames),
                kernel="operator",
                )

    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    return prim.Subscript(lp.symbolic.TaggedVariable(name, "operator_precomputed"), tuple(prim.Variable(i) for i in quadrature_inames(element)))


def define_quadrature_position(name, index):
    qps_per_dir = quadrature_points_per_direction()
    qp_bound = qps_per_dir[index]
    expression = Subscript(Variable(name_oned_quadrature_points(qp_bound)),
                           (Variable(quadrature_inames()[index]),))
    instruction(expression=expression,
                assignee=Subscript(Variable(name), (index,)),
                forced_iname_deps=frozenset(quadrature_inames(element)),
                forced_iname_deps_is_final=True,
                tags=frozenset({"quad"}),
                )


@backend(interface="quad_pos", name="sumfact")
def pymbolic_quadrature_position(index, visitor):
    # Return the non-precomputed version
    if not get_form_option("precompute_quadrature_info"):
        name = 'pos'
        temporary_variable(name, shape=(local_dimension(),), shape_impl=("fv",))
        define_quadrature_position(name, index)
        return prim.Subscript(prim.Variable(name), (index,))

    assert isinstance(index, int)
    dim = local_dimension()
    qps_per_dir = quadrature_points_per_direction()
    local_qps_per_dir = local_quadrature_points_per_direction()
    local_qps_per_dir_str = '_'.join(map(str, local_qps_per_dir))
    name = "quad_points_dim{}_num{}_dir{}".format(dim, local_qps_per_dir_str, index)

    loopy_class_member(name,
                       shape=local_qps_per_dir,
                       classtag="operator",
                       dim_tags=",".join(["f"] * dim),
                       managed=True,
                       potentially_vectorized=True,
                       )

    # Precompute it in the constructor
    inames = constructor_quadrature_inames(dim)
    assignee = prim.Subscript(prim.Variable(name), tuple(prim.Variable(i) for i in inames))
    expression = Subscript(Variable(name_oned_quadrature_points(qps_per_dir[index])),
                           (prim.Variable(inames[index])))
    instruction(assignee=assignee,
                expression=expression,
                within_inames=frozenset(inames),
                kernel="operator",
                )

    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    return prim.Subscript(lp.symbolic.TaggedVariable(name, "operator_precomputed"), tuple(prim.Variable(i) for i in quadrature_inames(element)))


@backend(interface="qp_in_cell", name="sumfact")
def pymbolic_quadrature_position_in_cell(restriction):
    from dune.perftool.pdelab.geometry import to_cell_coordinates
    return to_cell_coordinates(pymbolic_quadrature_position(), restriction)
